package locker

import (
	"context"

	"git.woa.com/khaos/eros/app/khaos-oss/internal/dao"
	"git.woa.com/khaos/eros/app/khaos-oss/internal/model"
	"git.woa.com/khaos/eros/app/khaos-oss/internal/model/do"
	"git.woa.com/khaos/eros/app/khaos-oss/internal/model/entity"
	"git.woa.com/khaos/eros/app/khaos-oss/internal/service"
	"git.woa.com/khaos/eros/utility/logger"
	"github.com/gogf/gf/v2/os/gtime"
)

type sLocker struct{}

func init() {
	service.RegisterLocker(New())
}

func New() *sLocker {
	return &sLocker{}
}

// LockFunc locks a database record using a unique name, an expired duration and a closer function.
// It will Unlock the lock automatically if it retrieved the lock with success and the closer function `f` done.
//
// Also see Lock.
func (s *sLocker) LockFunc(ctx context.Context, in model.LockerLockInput, f func() error) (ok bool, err error) {
	instance, ok, err := s.Lock(ctx, in)
	if err != nil {
		return false, err
	}
	if !ok {
		return false, nil
	}
	defer func() {
		if err = instance.Unlock(ctx); err != nil {
			logger.Error(ctx, err)
		}
	}()
	if err = f(); err != nil {
		return true, err
	}
	return true, nil
}

// Lock locks a database record using a unique name and an expired duration.
// Only one request can retrieve the lock at the same time.
//
// The returned parameter `instance` is the locker instance that is used for Unlock later after the locking business done.
// If you forget call Unlock function after retrieving the lock instance, it will automatically Unlock after `Duration`
// which is specified in model.LockerLockInput.
//
// The returned parameter `ok` specified whether retrieving the lock with success or not. It returns false if another
// request is holding the same lock.
func (s *sLocker) Lock(ctx context.Context, in model.LockerLockInput) (instance service.ILockerInstance, ok bool, err error) {
	var locker *entity.Locker
	err = dao.Locker.Ctx(ctx).Where(do.Locker{
		Name: in.Name,
	}).Scan(&locker)
	if err != nil {
		return nil, false, err
	}
	if locker == nil {
		result, err := dao.Locker.Ctx(ctx).Data(do.Locker{
			Name:      in.Name,
			Reason:    in.Reason,
			CreatedAt: gtime.Now(),
			ExpiredAt: gtime.Now().Add(in.Duration),
		}).InsertIgnore()
		if err != nil {
			return nil, false, err
		}
		insertId, err := result.LastInsertId()
		if err != nil {
			return nil, false, err
		}
		if insertId <= 0 {
			return nil, false, nil
		}
		return &sLockerInstance{Id: insertId}, true, nil
	}
	// Not expired.
	if gtime.Now().Before(locker.ExpiredAt) {
		return nil, false, nil
	}
	// Update the locker.
	affected, err := dao.Locker.Ctx(ctx).Data(do.Locker{
		Name:      in.Name,
		Reason:    in.Reason,
		CreatedAt: gtime.Now(),
		ExpiredAt: gtime.Now().Add(in.Duration),
	}).Where(locker).UpdateAndGetAffected()
	if err != nil {
		return nil, false, err
	}
	if affected <= 0 {
		return nil, false, nil
	}
	return &sLockerInstance{Id: int64(locker.Id)}, true, nil
}
